// Copyright (C) 1999-2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of Common C++.
// 
// The exception is that, if you link the Common C++ library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the Common C++ library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name Common C++.  If you copy code from other releases into a copy of
// Common C++, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for Common C++, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include "config.h"
#include "macros.h"
#include "thread.h"

#ifdef	_SYSV_SEMAPHORES

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <errno.h>

Semaphore::Semaphore(size_t resource)
{
	union semun arg;

	if((_semaphore = semget(IPC_PRIVATE, 1, 0644 | IPC_CREAT)) == -1)
	{
		if(getException() == THROW_OBJECT)
			throw(this);
		return;
	}

	arg.val = resource;
	if(semctl(_semaphore, 0, SETVAL, arg) == -1)
	{
		if(getException() == THROW_OBJECT)
			throw(this);
		return;
	}
}

Semaphore::~Semaphore()
{
	semctl(_semaphore, 0, IPC_RMID);
}

void Semaphore::Wait(void)
{
	struct sembuf ops[] = {{0, -1, 0}};
	
	semop(_semaphore, ops, 1);
}

bool Semaphore::TryWait(void)
{
	struct sembuf ops[] = {{0, -1, IPC_NOWAIT}};
	
	return (semop(_semaphore, ops, 1) == EAGAIN) ? false : true;
}

void Semaphore::Post(void)
{
	struct sembuf ops[] = {{0, 1, 0}};

	semop(_semaphore, ops, 1);
}

int Semaphore::getValue(void)
{
	return semctl(_semaphore,0,GETVAL);
}

#else

Semaphore::Semaphore(size_t resource) 
{
	if(sem_init(&_semaphore, 0, resource))
		if(getException() == THROW_OBJECT)
			throw(this);
}

#ifdef	__linux__

Semaphore::~Semaphore()
{
	sem_destroy(&_semaphore);
}

void Semaphore::Wait(void)
{
	sem_wait(&_semaphore);
}

void Semaphore::Post(void)
{
	sem_post(&_semaphore);
}

bool Semaphore::TryWait(void)
{
	return (sem_trywait(&_semaphore) == 0) ? true : false;
}

int Semaphore::getValue(void)
{
	int value;
	sem_getvalue(&_semaphore, &value);
	return value;
}

#endif
#endif

/** EMACS **
 * Local variables:
 * mode: c++
 * c-basic-offset: 8
 * End:
 */
