// Copyright (C) 1999-2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of Common C++.
// 
// The exception is that, if you link the Common C++ library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the Common C++ library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name Common C++.  If you copy code from other releases into a copy of
// Common C++, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for Common C++, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#ifndef	__CCXX_THREAD_H__
#define	__CCXX_THREAD_H__

#include <setjmp.h>

#ifndef	__CCXX_CONFIG_H__
#include <cc++/config.h>
#endif

#if _MSC_VER > 1000
#pragma once
#endif

typedef	DWORD	cctid_t;
typedef DWORD   timeout_t;

typedef enum {
	THROW_NOTHING, 
	THROW_OBJECT, 
	THROW_EXCEPTION
}		throw_t;

#define TIMEOUT_INF ~((timeout_t) 0)

#define	MAX_SEM_VALUE	1000000

#define	ENTER_CRITICAL	EnterMutex();
#define	LEAVE_CRITICAL	LeaveMutex();

typedef	enum {
	THREAD_CANCEL_IMMEDIATE = 1,
	THREAD_CANCEL_DEFERRED,
	THREAD_CANCEL_DEFAULT = THREAD_CANCEL_DEFERRED,
	THREAD_CANCEL_DISABLED,
	THREAD_CANCEL_INITIAL
} thread_cancel_t;

class Thread;
#define sleep(x) delay(x * 1000l)
#undef Yield

#if defined(__MINGW32__) || defined(__CYGWIN32__)

#include <Windows32/CommonFunctions.h>
#else
__declspec(dllimport) long __stdcall InterlockedIncrement(long *);
__declspec(dllimport) long __stdcall InterlockedDecrement(long *);
__declspec(dllimport) long __stdcall InterlockedExchange(long *, long);
__declspec(dllimport) long __stdcall InterlockedExchangeAdd(long *, long);

#endif

BOOL	__EXPORT APIENTRY DllMain(HANDLE hInstance, DWORD dwReason, LPVOID lpReserved);
Thread	__EXPORT *getThread(void);
DWORD	__EXPORT waitThread(HANDLE hRef, timeout_t timeout);
void	__EXPORT yield(void);
void	__EXPORT delay(timeout_t timeout);
void	__EXPORT setException(throw_t mode);
throw_t	__EXPORT getException(void);

class __EXPORT Mutex
{
private:
	HANDLE mutex;

public:
	Mutex();
	~Mutex(void)
		{::CloseHandle(mutex);};

	bool TryEnterMutex(void)
		{return (waitThread(mutex, 0) == WAIT_OBJECT_0);};
	void EnterMutex(void)
		{waitThread(mutex, INFINITE);};
	void LeaveMutex(void);

	friend inline void operator ++(Mutex &m)
		{m.EnterMutex();};

	friend inline void operator --(Mutex &m)
		{m.LeaveMutex();};
};

// R/W locker, handled as simple lock now
class __EXPORT ThreadLock
{
private:
  Mutex mutex;
public:
	ThreadLock();
	~ThreadLock();
	void ReadLock(void);
	void WriteLock(void);
	bool TryReadLock(void);
	bool TryWriteLock(void);
	void Unlock(void);
};

class __EXPORT MutexCounter : public Mutex 
{
private:
	int counter;

public:
	MutexCounter();
	MutexCounter(int initial);

	friend int operator++(MutexCounter &);
	friend int operator--(MutexCounter &);
};

class __EXPORT AtomicCounter
{
private:
	long atomic;

public:
	inline AtomicCounter()
		{atomic = 0;};

	inline AtomicCounter(int value)
		{atomic = value;};

	inline int operator++(void)
		{return InterlockedIncrement(&atomic);};

	inline int operator--(void)
		{return InterlockedDecrement(&atomic);};

	int operator+=(int change);

	int operator-=(int change);

	inline int operator+(int change)
		{return atomic + change;};

	inline int operator-(int change)
		{return atomic - change;};
	
	inline int operator=(int value)
		{return InterlockedExchange(&atomic, value);};

	inline bool operator!(void)
		{return (atomic == 0) ? true : false;};

	inline operator int()
		{return atomic;};
};

// FIXME: implement Conditional class

class __EXPORT Event 
{
private:
	HANDLE cond;

public:
	Event();

	~Event()
		{CloseHandle(cond);};

	inline void Reset(void)
		{ResetEvent(cond);};

	inline void Signal(void)
		{SetEvent(cond);};

	inline bool Wait(void)
		{return (waitThread(cond, INFINITE) == WAIT_OBJECT_0);};

	inline bool Wait(timeout_t timer)
		{return (waitThread(cond, timer) == WAIT_OBJECT_0);};

	friend inline void reset(Event &ev)
		{ResetEvent(ev.cond);};

	friend inline void signal(Event &ev)
		{SetEvent(ev.cond);};

	friend inline void wait(Event &ev)
		{waitThread(ev.cond, INFINITE);};

	friend inline void wait(Event &ev, timeout_t timer)
		{waitThread(ev.cond, timer);};
};

class __EXPORT Semaphore 
{
private:
	HANDLE	semObject;

public:
	~Semaphore()
		{CloseHandle(semObject);};

	Semaphore(size_t count = 0)
		{semObject = CreateSemaphore((LPSECURITY_ATTRIBUTES)NULL, count, MAX_SEM_VALUE, (LPCTSTR)NULL);};

	bool TryWait(void)
		{return waitThread(semObject, 0) == WAIT_OBJECT_0;};

	void Post(void)
		{ReleaseSemaphore(semObject, 1, (LPLONG)NULL);};

	void Wait(void)
		{waitThread(semObject, INFINITE);};

  // FIXME: how implement getValue for posix compatibility ?
};


/**
 * The buffer class represents an IPC service that is built upon a buffer
 * of fixed capacity that can be used to transfer objects between one or
 * more producer and consumer threads.  Producer threads post objects
 * into the buffer, and consumer threads wait for and receive objects from
 * the buffer.  Semaphores are used to to block the buffer from overflowing
 * and indicate when there is data available, and mutexes are used to protect
 * multiple consumers and producer threads from stepping over each other.
 * 
 * The buffer class is an abstract class in that the actual data being
 * buffered is not directly specified within the buffer class itself.  The
 * buffer class should be used as a base class for a class that actually
 * impliments buffering and which may be aware of the data types actually
 * are being buffered.  A template class could be created based on buffer
 * for this purpose.  Another possibility is to create a class derived
 * from both Thread and Buffer which can be used to implement message passing
 * threads.
 * 
 * @author David Sugar <dyfet@ostel.com>
 * @short Producer/Consumer buffer for use between threads.
 */
class __EXPORT Buffer 
{
private:
	Mutex lock_head, lock_tail;
	Semaphore size_head, size_tail;
	size_t _size;
	size_t _used;

protected:
	/**
	 * Invoke derived class buffer peeking method.
	 * @return size of object found.
	 * @param buf pointer to copy contents of head of buffer to.
	 */
	virtual int OnPeek(void *buf) = 0;
	/**
	 * Invoke derived class object request from buffer.
	 * @return size of object returned.
	 * @param buf pointer to hold object returned from the buffer.
	 */
	virtual int OnWait(void *buf) = 0;
	/**
	 * Invoke derived class posting of object to buffer.
	 * @return size of object posted.
	 * @param buf pointer to object being posted to the buffer.
	 */
	virtual int OnPost(void *buf) = 0;
public:
	/**
	 * Create a buffer object of known capacity.
	 * @param capcity is the integer capacity of the buffer.
	 */
	Buffer(size_t capacity);
	/**
	 * In derived functions, may be used to free the actual memory
	 * used to hold buffered data.
	 */
	virtual ~Buffer()
		{return;};

	/**
	 * Return the capacity of the buffer as specified at creation.
	 * @return size of buffer.
	 */
	inline size_t getSize(void)
		{return _size;};
	
	/**
	 * Return the current capacity in use for the buffer.  Free space
	 * is technically getSize() - getUsed().
	 * @return integer used capacity of the buffer.
	 * @see #getSize
	 */
	inline size_t getUsed(void)
		{return _used;};

	/**
	 * Let one or more threads wait for an object to become available
	 * in the buffer.  The waiting thread(s) will wait forever if no
	 * object is ever placed into the buffer.
	 * 
	 * @return size of object passed by buffer in bytes.
	 * @param buf pointer to store object retrieved from the buffer.
	 */
	int Wait(void *buf);

	/**
	 * Post an object into the buffer and enable a waiting thread to
	 * receive it.
	 * 
	 * @return size of object posted in bytes.
	 * @param buf pointer to object to store in the buffer.
	 */
	int Post(void *buf);

	/**
	 * Peek at the current content (first object) in the buffer.
	 * 
	 * @return size of object in the buffer.
	 * @param buf pointer to store object found in the buffer.
	 */
	int Peek(void *buf);

	/**
	 * New virtual to test if buffer is a valid object.
	 * @return true if object is valid.
	 */
	virtual bool isValid(void)
		{return true;};
};

/**
 * A buffer class that holds a known capacity of fixed sized objects defined
 * during creation.
 * 
 * @author David Sugar <dyfet@ostel.com>
 * @short producer/consumer buffer for fixed size objects.
 */
class __EXPORT FixedBuffer : public Buffer 
{
private:
	char *buf, *head, *tail;
	size_t objsize;

protected:
	/**
	 * Return the first object in the buffer.
	 * @return predefined size of this buffers objects.
	 * @param buf pointer to copy contents of head of buffer to.
	 */
	int OnPeek(void *buf);

	/**
	 * Wait for and return a fixed object in the buffer.
	 * @return predefined size of this buffers objects.
	 * @param buf pointer to hold object returned from the buffer.
	 */
	int OnWait(void *buf);

	/**
	 * Post an object of the appropriate size into the buffer.
	 * @return predefined size of this buffers objects.
	 * @param buf pointer to data to copy into the buffer.
	 */
	int OnPost(void *buf);	

public:
	/**
	 * Create a buffer of known capacity for objects of a specified
	 * size.
	 * 
	 * @param capacity of the buffer.
	 * @param objsize for each object held in the buffer.
	 */
	FixedBuffer(size_t capacity, size_t objsize);

	/**
	 * Create a copy of an existing fixed size buffer and duplicate
	 * it's contents.
	 * 
	 * @param fb existing FixedBuffer object.
	 */
	FixedBuffer(const FixedBuffer &fb);

	/**
	 * Destroy the fixed buffer and free the memory used to store objects.
	 */
	~FixedBuffer();

	FixedBuffer &operator=(const FixedBuffer &fb);

	bool isValid(void);
};

// FIXME: too different from posix implementation
class Thread 
{
private:
	static Thread *_main;
	Thread *_parent;
	DWORD _tid;
	HANDLE _cancellation;
	thread_cancel_t	_cancel;
	jmp_buf _env;
	Semaphore *_start;
	bool _active;
	throw_t _throw;

	static unsigned __stdcall Execute(Thread *th);

protected:
	HANDLE	_hThread;

	__MEMBER_EXPORT virtual void Initial(void)
		{return;};

	__MEMBER_EXPORT virtual void Final(void)
		{return;};

	virtual void Run(void) = 0;

	inline void Exit(void)
		{longjmp(_env, 1);};

	inline void clrParent(void)
		{_parent = NULL;};

	__MEMBER_EXPORT void setCancel(thread_cancel_t cancel);
	__MEMBER_EXPORT void Yield(void);
	__MEMBER_EXPORT void Terminate(void);
	__MEMBER_EXPORT void Sleep(timeout_t timeout);
	__MEMBER_EXPORT DWORD WaitHandle(HANDLE obj, timeout_t timeout);

public:
	inline Thread *getParent(void)
		{return _parent;};

	inline thread_cancel_t getCancel(void)
		{return _cancel;};

	inline bool isCancelled(void)
		{return waitThread(_cancellation, 0) == WAIT_OBJECT_0; };

	inline bool isCancelled(timeout_t timer)
		{return waitThread(_cancellation, timer) == WAIT_OBJECT_0; };

	inline bool isRunning(void)
		{return (_tid != 0) ? true : false;};

	inline bool isThread(void)
		{return ((_tid == GetCurrentThreadId())) ? true : false;};

	__MEMBER_EXPORT Thread(Semaphore *start = (Semaphore *)NULL, int pri = 0, unsigned stack = 0);
	__MEMBER_EXPORT Thread(bool flag);

	__MEMBER_EXPORT int Start(Semaphore *sem = (Semaphore *)NULL);

	__MEMBER_EXPORT virtual ~Thread()
		{Terminate();};

	__MEMBER_EXPORT virtual void *getExtended(void)
		{return NULL;};

	__MEMBER_EXPORT virtual void Notify(Thread *th)
		{return;};

	__MEMBER_EXPORT	void Suspend(void);
	__MEMBER_EXPORT	void Resume(void);

	friend inline int start(Thread &th, Semaphore *sem = (Semaphore *)NULL)
		{return th.Start(sem);};

	friend inline void operator++(Thread &th)
		{th._start->Post();};

	friend DWORD __EXPORT waitThread(HANDLE hRef, timeout_t timeout);
	friend void __EXPORT delay(timeout_t timeout);
	friend void __EXPORT yield(void);

	friend throw_t __EXPORT getException(void);
	friend void __EXPORT setException(throw_t mode);
};

class __EXPORT ThreadKey 
{
private:
	DWORD	key;

public:
	ThreadKey();
	~ThreadKey();

	void *getKey(void);
	void setKey(void *key);

	friend inline void *getKey(ThreadKey &th)
		{return th.getKey();};

	friend inline void setKey(ThreadKey &th, void *ptr)
		{th.setKey(ptr);};
};

class __EXPORT TimerPort
{
	DWORD timer;
	bool active;

public:
	TimerPort();
	void setTimer(timeout_t timeout = 0);
	void incTimer(timeout_t timeout);
	void endTimer(void);
	timeout_t getTimer(void);
	timeout_t getElapsed(void);
};

inline int get(Buffer &b, void *o)
	{return b.Wait(o);};

inline int put(Buffer &b, void *o)
	{return b.Post(o);};

inline int peek(Buffer &b, void *o)
	{return b.Peek(o);};

inline void operator++(Semaphore &s)
	{s.Post();};

inline void operator--(Semaphore &s)
	{s.Wait();};

#endif
/** EMACS **
 * Local variables:
 * mode: c++
 * c-basic-offset: 8
 * End:
 */
